package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.block.entity.worker.BlockEntityMatteryWorker
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJob
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJobStatus
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.MatteryMachineEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.MatterHandlerCapability
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.matter.MatterRegistry
import ru.dbotthepony.mc.otm.menu.MatterDecomposerMenu
import ru.dbotthepony.mc.otm.set

class BlockEntityMatterDecomposer(pos: BlockPos, state: BlockState)
	: BlockEntityMatteryWorker(Registry.BlockEntities.MATTER_DECOMPOSER, pos, state), IMatterGraphNode {

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_decomposer")
		private val BASE_CONSUMPTION = Fraction(240)
	}

	init {
		energy = MatteryMachineEnergyStorage(
			this,
			MatteryMachineEnergyStorage.MachineType.WORKER,
			Fraction(400000),
			Fraction(2000),
			Fraction(2000)
		)
	}

	private var valid = true
	private val node = Graph6Node<IMatterGraphNode>(this)

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	@JvmField
	val matter = MatterHandlerCapability(this::setChangedLight, IMatterHandler.MatterDirection.EXTRACT, Fraction("20"))

	private var resolverMatter = LazyOptional.of { matter }
	private var resolverNode = LazyOptional.of { this }

	// вход, выход
	@JvmField
	val itemContainer = MatteryContainer(this::setChangedLight, 2)

	private val itemHandler = LazyOptional.of<IItemHandler> {
		itemContainer.handler(
			{ slot: Int, stack: ItemStack? -> slot == 0 && MatterRegistry.canDecompose(stack) },
			{ slot: Int, amount: Int, stack: ItemStack? -> slot == 1 })
	}

	override fun getDefaultDisplayName(): Component {
		return MACHINE_NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterDecomposerMenu(containerID, inventory, this)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["work_slots"] = itemContainer.serializeNBT()
		nbt["matter_capability"] = matter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("matter_capability") {
			if (it is CompoundTag)
				matter.deserializeNBT(it)
		}

		nbt.ifHas("work_slots") {
			itemContainer.deserializeNBT(it)
		}
	}

	override fun reviveCaps() {
		valid = true
		super.reviveCaps()
		resolverMatter = LazyOptional.of { matter }
	}

	override fun invalidateCaps() {
		valid = false
		super.invalidateCaps()
		resolverMatter.invalidate()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap === MatteryCapability.MATTER) return resolverMatter.cast()
			if (cap === MatteryCapability.MATTER_NODE) return resolverNode.cast()
			if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) return itemHandler.cast()
		}

		return super.getCapability(cap, side)
	}

	override fun onJobFinish(job: MachineJob): MachineJobStatus {
		matter.receiveMatterInner(MatterRegistry.getMatterValue(job.stack()), false)
		return MachineJobStatus()
	}

	override fun getNextJob(): MachineJob? {
		val stack = itemContainer.getItem(0)

		if (!stack.isEmpty) {
			val copy = stack.copy()
			copy.count = 1

			if (MatterRegistry.canDecompose(copy)) {
				val value = MatterRegistry.getMatterValue(copy)

				if (value != Fraction.ZERO && matter.canReceiveAll(value)) {
					stack.shrink(1)
					return MachineJob(copy, value.toDouble() * 12500.0)
				}
			}
		}

		return null
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	override fun getMatterHandler(): IMatterHandler {
		return matter
	}

	override fun getBaseConsumption(): Fraction {
		return BASE_CONSUMPTION
	}

	fun tick() {
		batteryChargeLoop();
		workerLoop();

		val grid = node.graph as MatterNetworkGraph?

		if (!matter.storedMatter.isZero() && grid != null) {
			val diff = matter.extractMatterInner(matter.getStoredMatter(), true)
			val diff2 = grid.receiveMatter(diff, true)

			matter.extractMatterInner(diff2, false)
			grid.receiveMatter(diff2, false)
		}
	}
}
