package ru.dbotthepony.mc.otm.container

import net.minecraft.MethodsReturnNonnullByDefault
import javax.annotation.ParametersAreNonnullByDefault
import net.minecraft.world.item.ItemStack
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.world.Container
import kotlin.jvm.JvmOverloads
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set
import java.util.*
import java.util.function.Consumer

open class MatteryContainer(val watcher: Runnable, private val size: Int) : Container, Iterable<ItemStack> {
	constructor(watcher: BlockEntity, size: Int) : this(watcher::setChanged, size)
	// constructor(watcher: BlockEntityMattery, size: Int) : this(watcher::setChangedLight, size)

	private var ignoreChangeNotifications = 0
	@JvmField protected val slots: Array<ItemStack> = Array(size) { ItemStack.EMPTY }
	private val trackedSlots: Array<ItemStack> = Array(size) { ItemStack.EMPTY }

	final override fun getContainerSize() = size

	fun startIgnoreUpdates() {
		if (++ignoreChangeNotifications == 1) {
			startedIgnoringUpdates()
		}
	}

	fun stopIgnoreUpdates() {
		if (--ignoreChangeNotifications == 0) {
			stoppedIgnoringUpdates()
		}
	}

	protected open fun startedIgnoringUpdates() {}
	protected open fun stoppedIgnoringUpdates() {}

	fun deserializeNBT(tag: CompoundTag?) {
		startIgnoreUpdates()

		for (i in 0 until size) {
			setItem(i, ItemStack.EMPTY)
		}

		if (tag == null) {
			stopIgnoreUpdates()
			return
		}

		// нам не интересен размер
		tag.ifHas("items", ListTag::class.java) {
			for (i in 0 until Math.min(it.size, size)) {
				setItem(i, ItemStack.of(it[i] as CompoundTag))
			}
		}

		stopIgnoreUpdates()
		setChanged()
	}

	fun deserializeNBT(tag: Tag?) {
		if (tag is CompoundTag) {
			deserializeNBT(tag)
		}
	}

	fun stacks(): ContainerIteratorItemStack {
		return ContainerIteratorItemStack(this)
	}

	fun forEachNonEmpty(consumer: (ItemStack) -> Unit) {
		for (item in slots) {
			if (!item.isEmpty) {
				consumer(item)
			}
		}
	}

	fun forEachNonEmpty(consumer: (ItemStack, Int) -> Unit) {
		for (i in 0 until size) {
			val item = slots[i]

			if (!item.isEmpty) {
				consumer(item, i)
			}
		}
	}

	fun <T> capabilityIterator(cap: Capability<T>): ContainerIteratorCapability<T> {
		return ContainerIteratorCapability(this, cap)
	}

	fun <T> consumeCapability(cap: Capability<T>, consumer: Consumer<T>) {
		capabilityIterator(cap).consume(consumer)
	}

	open fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
		if (ignoreChangeNotifications == 0) watcher.run()
	}

	fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()
		tag["size"] = size

		val list = ListTag()
		tag["items"] = list

		for (i in 0 until size)
			list.add(this[i].serializeNBT())

		return tag
	}

	fun hasEmptySlot(): Boolean {
		for (i in 0 until size) {
			if (this[i].isEmpty) {
				return true
			}
		}

		return false
	}

	protected var handler: MatteryContainerHandler? = null

	fun handler(
		insert_validator: (Int, ItemStack) -> Boolean,
		extract_validator: (Int, Int, ItemStack) -> Boolean
	): MatteryContainerHandler {
		return handler ?: MatteryContainerHandler(this, insert_validator, extract_validator).also { handler = it }
	}

	fun handler(insert_validator: (Int, ItemStack) -> Boolean): MatteryContainerHandler {
		return handler ?: MatteryContainerHandler(this, insert_validator).also { handler = it }
	}

	fun handler(): MatteryContainerHandler {
		return handler ?: MatteryContainerHandler(this).also { handler = it }
	}

	open fun getMaxStackSize(slot: Int) = maxStackSize

	@JvmOverloads
	fun addItem(stack: ItemStack, start: Int = 0, end: Int = size, simulate: Boolean = false): ItemStack {
		if (stack.isEmpty || start < 0 || end > size || start >= end)
			return ItemStack.EMPTY

		val copy = stack.copy()

		// двигаем в одинаковые слоты
		for (slot in start until end) {
			if (ItemStack.isSameItemSameTags(slots[slot], copy)) {
				val slotStack = slots[slot]
				val slotLimit = Math.min(getMaxStackSize(slot), slotStack.maxStackSize)

				if (slotStack.count < slotLimit) {
					val newCount = Math.min(slotStack.count + copy.count, slotLimit)
					val diff = newCount - slotStack.count

					if (!simulate) {
						val old = slotStack.copy()
						slotStack.count = newCount
						trackedSlots[slot] = slotStack.copy()
						setChanged(slot, slotStack, old)
					}

					copy.shrink(diff)

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// двигаем в пустые слоты
		for (slot in start until end) {
			if (slots[slot].isEmpty) {
				val diff = Math.min(copy.count, Math.min(getMaxStackSize(slot), copy.maxStackSize))

				if (!simulate) {
					val copyToPut = copy.copy()
					copyToPut.count = diff
					this[slot] = copyToPut
				}

				copy.shrink(diff)

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun addItem(stack: ItemStack, simulate: Boolean): ItemStack {
		return addItem(stack, 0, size, simulate)
	}

	@JvmOverloads
	fun fullyAddItem(stack: ItemStack, start: Int = 0, end: Int = size): Boolean {
		if (!addItem(stack, start, end, true).isEmpty)
			return false

		return addItem(stack, start, end, false).isEmpty
	}

	override fun isEmpty(): Boolean {
		for (stack in slots)
			if (!stack.isEmpty)
				return false

		return true
	}

	operator fun get(slot: Int) = getItem(slot)
	operator fun set(slot: Int, stack: ItemStack) = setItem(slot, stack)

	operator fun contains(other: ItemStack): Boolean {
		for (i in 0 until size) {
			if (ItemStack.isSameItemSameTags(this[i], other)) {
				return true
			}
		}

		return false
	}

	final override fun getItem(slot: Int): ItemStack {
		val item = slots[slot]

		if (item.isEmpty)
			return ItemStack.EMPTY
		else
			return item
	}

	final override fun removeItem(slot: Int, amount: Int): ItemStack {
		if (amount <= 0 || slot < 0 || slot >= size || slots[slot].isEmpty)
			return ItemStack.EMPTY

		val old = slots[slot].copy()
		val split = slots[slot].split(amount)
		trackedSlots[slot] = slots[slot].copy()
		setChanged(slot, if (slots[slot].isEmpty) ItemStack.EMPTY else slots[slot], old)

		return split
	}

	final override fun removeItemNoUpdate(slot: Int): ItemStack {
		val old = slots[slot]
		slots[slot] = ItemStack.EMPTY
		trackedSlots[slot] = ItemStack.EMPTY
		return old
	}

	final override fun setItem(slot: Int, stack: ItemStack) {
		if (slots[slot].isEmpty && stack.isEmpty || stack === slots[slot])
			return

		val old = slots[slot]
		slots[slot] = stack
		trackedSlots[slot] = stack.copy()
		setChanged(slot, stack, old)
	}

	final override fun setChanged() {
		for (slot in 0 until size) {
			if (!ItemStack.isSameItemSameTags(slots[slot], trackedSlots[slot])) {
				setChanged(slot, slots[slot], trackedSlots[slot])
				trackedSlots[slot] = slots[slot].copy()
				// mojang соси))0)0))0)))))0)
			}
		}
	}

	override fun stillValid(player: Player): Boolean {
		return true
	}

	final override fun clearContent() {
		for (slot in 0 until size) {
			if (!slots[slot].isEmpty) {
				val old = slots[slot]
				slots[slot] = ItemStack.EMPTY
				setChanged(slot, ItemStack.EMPTY, old)
			}
		}

		Arrays.fill(trackedSlots, ItemStack.EMPTY)
	}

	final override fun iterator(): Iterator<ItemStack> {
		return slots.iterator()
	}
}
