package ru.dbotthepony.mc.otm.tests

import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import ru.dbotthepony.mc.otm.core.Fraction
import java.lang.Math.floor
import java.math.BigDecimal
import java.math.BigInteger
import kotlin.random.Random

object FractionTests {
	@Test
	@DisplayName("Fraction declaration")
	fun declaration() {
		println("BigInteger 1/1 == ${Fraction(BigInteger.valueOf(1), BigInteger.valueOf(1))}")
		println("BigDecimal 1 == ${Fraction(BigDecimal("1"))}")

		var one = BigDecimal("1.00")

		println("Unscaled ${one.unscaledValue()} with scale ${one.scale()} of $one")
		println("BigDecimal $one == ${Fraction(one)}")
		assert(Fraction(one).compareTo(Fraction.ONE) == 0)

		one = BigDecimal("1.0000")

		println("Unscaled ${one.unscaledValue()} with scale ${one.scale()} of $one")
		println("BigDecimal $one == ${Fraction(one)}")
		assert(Fraction(one).compareTo(Fraction.ONE) == 0)

		println("1/2 == ${Fraction(1, 2)}")
		println("-1/2 == ${Fraction(-1, 2)}")
		println("1/-2 == ${Fraction(1, -2)}")
		println("-1/-2 == ${Fraction(-1, -2)}")

		println("canonical 1/2 == ${Fraction(1, 2).canonize()}")
		println("canonical -1/2 == ${Fraction(-1, 2).canonize()}")
		println("canonical 1/-2 == ${Fraction(1, -2).canonize()}")
		println("canonical -1/-2 == ${Fraction(-1, -2).canonize()}")
	}

	@Test
	@DisplayName("Fraction NaN behavior")
	fun nan() {
		assert((Fraction.NaN + Fraction.ONE).isNaN())
		assert((Fraction.NaN - Fraction.ONE).isNaN())
		assert((Fraction.NaN / Fraction.ONE).isNaN())
		assert((Fraction.NaN * Fraction.ONE).isNaN())

		assert((Fraction.ONE + Fraction.NaN).isNaN())
		assert((Fraction.ONE - Fraction.NaN).isNaN())
		assert((Fraction.ONE / Fraction.NaN).isNaN())
		assert((Fraction.ONE * Fraction.NaN).isNaN())

		assert((Fraction.ONE / Fraction.ZERO).isNaN())
		assert((Fraction.ONE * Fraction(0, 0)).isNaN())
		assert((Fraction.ONE * Fraction(1, 0)).isNaN())

		assert(Fraction.NaN.toFloat().isNaN())
		assert(Fraction.NaN.toDouble().isNaN())

		assertThrows<ArithmeticException> {Fraction.NaN.toInt()}
		assertThrows<ArithmeticException> {Fraction.NaN.toLong()}
		assertThrows<ArithmeticException> {Fraction.NaN.wholePart()}

		assert(Fraction.fromByteArray(Fraction.NaN.toByteArray()).isNaN())
		assert(Fraction.NaN.percentage(Fraction.ONE).isNaN())
	}

	@Test
	@DisplayName("Fraction comparison")
	fun equality() {
		assert(Fraction(1).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0F).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1, 1).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0, 1.0).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0F, 1.0F).compareTo(Fraction.ONE) == 0)

		assert(Fraction(1, 2).compareTo(Fraction.HALF) == 0)
		assert(Fraction(1.0, 2.0).compareTo(Fraction.HALF) == 0)
		assert(Fraction(1.0F, 2.0F).compareTo(Fraction.HALF) == 0)

		assert(Fraction(-1, -2).compareTo(Fraction.HALF) == 0)
		assert(Fraction(-1.0, -2.0).compareTo(Fraction.HALF) == 0)
		assert(Fraction(-1.0F, -2.0F).compareTo(Fraction.HALF) == 0)

		assert(Fraction(-1, 2).compareTo(Fraction.HALF) != 0)
		assert(Fraction(-1.0, 2.0).compareTo(Fraction.HALF) != 0)
		assert(Fraction(-1.0F, 2.0F).compareTo(Fraction.HALF) != 0)

		assert(Fraction(1, 2).compareTo(Fraction.ONE) != 0)
		assert(Fraction(1.0, 2.0).compareTo(Fraction.ONE) != 0)
		assert(Fraction(1.0F, 2.0F).compareTo(Fraction.ONE) != 0)

		assert(Fraction(2).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0F).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2, 1).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0, 1.0).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0F, 1.0F).compareTo(Fraction.TWO) == 0)

		assert(Fraction(4, 3) > Fraction(5, 4))
		assert(Fraction(4, 4) < Fraction(5, 4))

		assert(Fraction(-15, 2) < Fraction(2, 4))
		assert(Fraction(-1, 2) < Fraction(2, 4))
		assert(Fraction(-15, -2) > Fraction(2, 4))
		assert(Fraction(-15, -2) > Fraction(-2, -4))
		assert(Fraction(-15, 2) < Fraction(-2, -4))
	}

	@Test
	@DisplayName("Fraction math")
	fun math() {
		assert((Fraction(1) / Fraction(1)) == Fraction(1))
		assert((Fraction(2) / Fraction(1)) == Fraction(2, 1))
		assert((Fraction(2) / Fraction(2)) == Fraction(1)) { Fraction(2) / Fraction(2) }

		assert((Fraction(4, 3) / Fraction(5, 4)) == (Fraction(4, 3) * Fraction(4, 5)))
		assert((Fraction(4, 3) + Fraction(5, 4)) == Fraction(31, 12))
	}

	@Test
	@DisplayName("Fraction compacting")
	fun compacting() {
		val frac = Fraction("2721280000000000000000000000000000000000000000000000000000000000000000000000000000000000000000", "25600000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000")

		println("Compacting $frac => ${frac + Fraction.ZERO}")
	}
	
	private val samples = arrayOf(
		Fraction(9475, 4729),
		Fraction(23535, 58723),
		Fraction(-4852, 6859),
		Fraction(-45623, -76849),
		Fraction(38494, -76849),
		Fraction(1043, -648),
	)

	private val samples2 = arrayOf(
		(9475.0 /  4729),
		(23535.0 /  58723),
		(-4852.0 /  6859),
		(-45623.0 / -76849),
		(38494.0 / -76849),
		(1043.0 / -648),
	)

	private val samples3 = arrayOf(
		BigDecimal((9475.0 /  4729).toString()),
		BigDecimal((23535.0 /  58723).toString()),
		BigDecimal((-4852.0 /  6859).toString()),
		BigDecimal((-45623.0 / -76849).toString()),
		BigDecimal((38494.0 / -76849).toString()),
		BigDecimal((1043.0 / -648).toString()),
	)

	@Test
	@DisplayName("Fraction performance measurement")
	fun performance() {
		val rand = java.util.Random()
		val blackHole = arrayOfNulls<Fraction>(100_000)
		val size = samples.size
		var time = System.currentTimeMillis()

		for (i in 0 until 100_000) {
			when (rand.nextInt(3)) {
				0 -> blackHole[i] = samples[rand.nextInt(size)] + samples[rand.nextInt(size)]
				1 -> blackHole[i] = samples[rand.nextInt(size)] - samples[rand.nextInt(size)]
				2 -> blackHole[i] = samples[rand.nextInt(size)] * samples[rand.nextInt(size)]
				// 3 -> blackHole[i] = samples[rand.nextInt(size)] / samples[rand.nextInt(size)]
			}
		}

		println("Mean time for Fraction operation is ~${System.currentTimeMillis() - time} ms per 100,000 ops")

		var sum = Fraction.ZERO

		// перемешаем чтоб оптимизатор не отбросил
		for (i in 0 until size) {
			sum += blackHole[i]!!
			blackHole[i] = blackHole[rand.nextInt(size)]
		}

		val blackHole2 = Array(100_000) {0.0}
		time = System.currentTimeMillis()

		for (i in 0 until 100_000) {
			when (rand.nextInt(3)) {
				0 -> blackHole2[i] = samples2[rand.nextInt(size)] + samples2[rand.nextInt(size)]
				1 -> blackHole2[i] = samples2[rand.nextInt(size)] - samples2[rand.nextInt(size)]
				2 -> blackHole2[i] = samples2[rand.nextInt(size)] * samples2[rand.nextInt(size)]
				// 3 -> blackHole2[i] = samples2[rand.nextInt(size)] / samples2[rand.nextInt(size)]
			}
		}

		println("Mean time for Double operation is ~${System.currentTimeMillis() - time} ms per 100,000 ops")

		val blackHole3 = arrayOfNulls<BigDecimal>(100_000)
		time = System.currentTimeMillis()

		for (i in 0 until 100_000) {
			when (rand.nextInt(3)) {
				0 -> blackHole3[i] = samples3[rand.nextInt(size)] + samples3[rand.nextInt(size)]
				1 -> blackHole3[i] = samples3[rand.nextInt(size)] - samples3[rand.nextInt(size)]
				2 -> blackHole3[i] = samples3[rand.nextInt(size)] * samples3[rand.nextInt(size)]
				// 3 -> blackHole2[i] = samples2[rand.nextInt(size)] / samples2[rand.nextInt(size)]
			}
		}

		println("Mean time for BigDecimal operation is ~${System.currentTimeMillis() - time} ms per 100,000 ops")

		var sum2 = 0.0

		// перемешаем чтоб оптимизатор не отбросил
		for (i in 0 until size) {
			sum2 += blackHole2[i]!!
			blackHole2[i] = blackHole2[rand.nextInt(size)]
		}

		var sum3 = 0.0

		// перемешаем чтоб оптимизатор не отбросил
		for (i in 0 until size) {
			sum3 += blackHole2[i]!!
			blackHole3[i] = blackHole3[rand.nextInt(size)]
		}

		println("$sum $sum2 $sum3")
	}

	@Test
	@DisplayName("Fraction serialization test")
	fun serialization() {
		var value = Fraction(1)
		var serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(4, 2)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(-4, 2)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(-4, -18)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("3407203485237459085739045724837543569234750927348902374590872345", "-57777772398450982374590230984532984")
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("320", "100")
		serialized = value.serializeNBT()
		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("324", "100")
		serialized = value.serializeNBT()
		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("328", "100")
		serialized = value.serializeNBT()
		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("332", "100")
		serialized = value.serializeNBT()
		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("336", "100")
		serialized = value.serializeNBT()
		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)
	}

	@Test
	@DisplayName("Fraction inaccurate representation")
	fun inaccurate() {
		var value = 1.1
		var frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 1.45
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = -1.0
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = -254.66
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 94.949494
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 1.0 / 3
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }
	}
}
